package bubble.cloud.email;

import bubble.dao.account.AccountDAO;
import bubble.dao.cloud.CloudServiceDAO;
import bubble.model.account.Account;
import bubble.model.cloud.CloudCredentials;
import bubble.model.cloud.CloudService;
import bubble.server.BubbleConfiguration;
import com.sendgrid.Method;
import com.sendgrid.Request;
import com.sendgrid.Response;
import com.sendgrid.SendGrid;
import lombok.*;
import org.springframework.beans.factory.annotation.Autowired;

import java.io.IOException;

import static bubble.cloud.storage.StorageCryptStream.MIN_DISTINCT_LENGTH;
import static bubble.cloud.storage.StorageCryptStream.MIN_KEY_LENGTH;
import static java.net.HttpURLConnection.HTTP_NO_CONTENT;
import static java.net.HttpURLConnection.HTTP_OK;
import static org.cobbzilla.util.daemon.ZillaRuntime.die;
import static org.cobbzilla.util.json.JsonUtil.EMPTY_JSON;
import static org.cobbzilla.util.json.JsonUtil.json;
import static org.cobbzilla.util.security.CryptoUtil.generatePassword;
import static org.cobbzilla.util.string.StringUtil.repeat;

public class SendgridSmtpEmailDriver extends SmtpEmailDriver {

    private static final String PARAM_PARENT_SERVICE = "parentService";

    @Autowired private AccountDAO accountDAO;
    @Autowired private CloudServiceDAO serviceDAO;

    @Override protected boolean isServiceCompatible() { return this.config.getHost().equals(SENDGRID_SMTP); }

    /**
     * Build username which will be used for Subuser created on SendGrid's service for specified Account's object.
     */
    @NonNull private String sgUsername(@NonNull final CloudService delegatedService) {
        return delegatedService.getShortId();
    }

    @Override @NonNull public CloudService setupDelegatedCloudService(@NonNull final BubbleConfiguration configuration,
                                                                      @NonNull final CloudService parentService,
                                                                      @NonNull final CloudService delegatedService) {
        final CloudCredentials parentCredentials = parentService.getCredentials();
        if (parentService.delegated() || !parentCredentials.getParam(PARAM_HOST).contains(".sendgrid.net")) {
            return super.setupDelegatedCloudService(configuration, parentService, delegatedService);
        }

        final SendGrid sg = new SendGrid(parentCredentials.getParam(PARAM_PASSWORD));
        final Account accountWithDelegate = accountDAO.findByUuid(delegatedService.getAccount());
        final String user = sgUsername(delegatedService);
        String password = generatePassword(MIN_KEY_LENGTH, MIN_DISTINCT_LENGTH);
        final CreateSubuserRequest data = new CreateSubuserRequest(user, accountWithDelegate.getEmail(), password,
                                                                   new String[]{});
        final Request req = new Request();
        req.setMethod(Method.POST);
        req.setEndpoint("subusers");
        req.setBody(json(data));

        final Response res;
        try {
            res = sg.api(req);
        } catch (IOException e) {
            return die("Cannot create SendGrid Subuser", e);
        }
        if (res.getStatusCode() != HTTP_OK) {
            return die("Wrong response when creating SendGrid Subuser: " + res.getStatusCode() + " : " + res.getBody());
        }

        delegatedService.setDelegated(null).setTemplate(false);
        delegatedService.getCredentials()
                        .setParam(PARAM_USER, user)
                        .setParam(PARAM_PASSWORD, password)
                        .setParam(PARAM_PARENT_SERVICE, parentService.getUuid());
        password = repeat("x", MIN_KEY_LENGTH); // Override password (in memory) just in case
        return delegatedService;
    }

    @Override public void postServiceDelete(@NonNull final CloudService service) {
        final String parentServiceUuid = service.getCredentials().getParam(PARAM_PARENT_SERVICE);
        if (parentServiceUuid == null) return;

        final CloudService parentService = serviceDAO.findByUuid(parentServiceUuid);
        if (parentService == null) return;

        final SendGrid sg = new SendGrid(parentService.getCredentials().getParam(PARAM_PASSWORD));
        final String sgUserToDelete = sgUsername(service);

        final Request req = new Request();
        req.setMethod(Method.DELETE);
        req.setEndpoint("subusers/" + sgUserToDelete);
        req.setBody(EMPTY_JSON);

        final Response res;
        try {
            res = sg.api(req);
        } catch (IOException e) {
            die("Cannot delete SendGrid Subuser " + sgUserToDelete, e);
            return;
        }
        if (res.getStatusCode() != HTTP_NO_CONTENT) {
            die("Wrong response when creating SendGrid Subuser: " + res.getStatusCode() + " : " + res.getBody());
        }
    }

    @AllArgsConstructor @NoArgsConstructor
    private class CreateSubuserRequest {
        @Getter @Setter private String username;
        @Getter @Setter private String email;
        @Getter @Setter private String password;
        @Getter @Setter private String[] ips;
    }
}
